from abc import abstractmethod

from django.contrib import admin, messages
from django.contrib.admin import ModelAdmin, helpers
from django.contrib.admin.utils import model_ngettext
from django.db.models import Model, QuerySet
from django.http import HttpRequest
from django.template.response import TemplateResponse

"""
Admin actions, for use in ModelAdmin.actions lists.

NB: In order to use mark_as_active and mark_as_inactive, the model must have
an `is_active` boolean field.
"""


def _set_is_active(modeladmin: ModelAdmin, request: HttpRequest, queryset: QuerySet, value: bool):
    queryset.update(is_active=value)
    modeladmin.message_user(
        request,
        'Marked %d object(s) as %s.' % (len(queryset), 'active' if value else 'inactive'),
        messages.SUCCESS,
    )


@admin.action(description='Mark selected %(verbose_name_plural)s as active')
def mark_as_active(modeladmin: ModelAdmin, request: HttpRequest, queryset: QuerySet):
    _set_is_active(modeladmin, request, queryset, True)


@admin.action(description='Mark selected %(verbose_name_plural)s as inactive')
def mark_as_inactive(modeladmin: ModelAdmin, request: HttpRequest, queryset: QuerySet):
    _set_is_active(modeladmin, request, queryset, False)


class IntermediatePageAction:
    """
    Generalization of actions displaying intermediate pages. Its template
    must reside in some admin template directory, and contain this field:

    >>> <input type="hidden" name="action" value="{{ action }}" />

    Usage is analoguous to how View.as_view() is used in urlconfs:

    >>> actions = [ActionClass.as_function()]
    """
    description: str
    template_name: str

    def __init__(self, modeladmin: ModelAdmin[Model], request: HttpRequest, queryset: QuerySet):
        self.request = request
        self.modeladmin = modeladmin
        self.queryset = queryset

    @classmethod
    def as_function(cls):
        @admin.display(description=cls.description)
        def action_func(modeladmin, request: HttpRequest, queryset: QuerySet):
            return cls(modeladmin, request, queryset).dispatch()
        action_func.__name__ = cls.__name__
        return action_func

    def dispatch(self):
        if self.request.POST.get("post"):
            self.post()
            return None
        return self.get()

    def get(self):
        return TemplateResponse(
            self.request,
            [
                "admin/{}/{}/{}".format(
                    self.modeladmin.model._meta.app_label,
                    self.modeladmin.model._meta.model_name,
                    self.template_name,
                ),
                "admin/{}/{}".format(self.modeladmin.model._meta.app_label, self.template_name),
                "admin/{}".format(self.template_name),
            ],
            self.get_context_data(),
        )

    def get_context_data(self, **kwargs):
        return {
            **self.modeladmin.admin_site.each_context(self.request),
            "queryset": self.queryset,
            "media": self.modeladmin.media,
            "opts": self.modeladmin.model._meta,
            "title": self.description,
            "action_checkbox_name": helpers.ACTION_CHECKBOX_NAME,
            "objects_name": model_ngettext(self.queryset),
            "action": self.__class__.__name__,
            **kwargs,
        }

    @abstractmethod
    def post(self):
        ...



import re
from datetime import date, datetime, timedelta
from typing import Any, Collection, Iterable
from urllib.parse import urljoin

from klaatu_python.utils import percent_rounded

from django import template
from django.conf import settings
from django.contrib.humanize.templatetags.humanize import NaturalTimeFormatter
from django.http.request import HttpRequest, QueryDict
from django.template.base import token_kwargs
from django.template.context import Context
from django.template.defaultfilters import stringfilter
from django.templatetags.static import static
from django.urls import reverse
from django.utils import timezone
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.timesince import timesince, timeuntil
from django.utils.translation import gettext_lazy, ngettext_lazy, override

from klaatu_django.utils import (
    _get_param_list,
    capitalize,
    natural_and_list,
    natural_or_list,
    relativedelta_rounded,
    render_modal,
    timedelta_formatter,
)

register = template.Library()


class EmailSection(template.Node):
    def __init__(self, nodelist: template.NodeList, **kwargs: template.base.FilterExpression):
        self.nodelist = nodelist
        self.kwargs = kwargs

    def render(self, context: template.Context) -> str:
        table_kwargs = {
            "role": "presentation",
            "class": "section",
            "cellspacing": "0",
            "cellpadding": "0",
        }
        table_args = ["%s=\"%s\"" % (k, v) for k, v in table_kwargs.items()]
        td_args = []
        for name, value in self.kwargs.items():
            td_args.append("%s=\"%s\"" % (name, value.resolve(context)))  # type: ignore
        return "<table %s><tr><td %s>%s</td></tr></table>" % (
            " ".join(table_args),
            " ".join(td_args),
            self.nodelist.render(context)
        )


class LinkNode(template.Node):
    def __init__(
        self,
        nodelist: template.NodeList,
        href: template.base.FilterExpression,
        **kwargs: template.base.FilterExpression,
    ):
        self.nodelist = nodelist
        self.href = href
        self.kwargs = kwargs

    def render(self, context: Context):
        href = self.href.resolve(context, ignore_failures=True)  # type: ignore
        kwargs = {
            key.replace("_", "-"): value.resolve(context, ignore_failures=True)  # type: ignore
            for key, value in self.kwargs.items()
        }
        root_url = getattr(settings, "ROOT_URL", "")
        if root_url and isinstance(href, str):
            is_external = bool(re.match("^https?://", href)) and not href.startswith(root_url)
        else:
            is_external = False
        if is_external:
            if "rel" in kwargs:
                if "external" not in kwargs["rel"].split(" "):
                    kwargs["rel"] += " external"
            else:
                kwargs["rel"] = "external"

        return format_html(
            '<a href="{}"{}{}>{}</a>',
            href,
            mark_safe(' target="_blank"') if is_external else " ",
            mark_safe("".join([f' {key}="{value}"' for key, value in kwargs.items()])),
            self.nodelist.render(context),
        )


class NaturalTimeShortFormatterMeta(type):
    def __new__(mcs, name, bases, dct):
        """
        The 'past' ones are identical to the originals, but I'm including them
        because the default Swedish translations are wrong.
        """
        klass = super().__new__(mcs, name, bases, dct)
        time_strings = getattr(klass, "time_strings", {})
        time_strings.update({
            'past-day': gettext_lazy('%(delta)s ago'),
            'past-hour': ngettext_lazy('an hour ago', '%(count)s hours ago', 'count'),
            'past-minute': ngettext_lazy('a minute ago', '%(count)s minutes ago', 'count'),
            'past-second': ngettext_lazy('a second ago', '%(count)s seconds ago', 'count'),
            'future-second': ngettext_lazy('in one second', 'in %(count)s seconds', 'count'),
            'future-minute': ngettext_lazy('in one minute', 'in %(count)s minutes', 'count'),
            'future-hour': ngettext_lazy('in one hour', 'in %(count)s hours', 'count'),
            'future-day': gettext_lazy('in %(delta)s'),
            "yesterday": gettext_lazy("yesterday"),
            "tomorrow": gettext_lazy("tomorrow"),
        })
        setattr(klass, "time_strings", time_strings)
        return klass


class NaturalTimeShortFormatter(NaturalTimeFormatter, metaclass=NaturalTimeShortFormatterMeta):
    """
    Ugly hack to only return the first part of the string when the timedelta
    >= 1 day (NaturalTimeFormatter returns "1 month, 3 days" etc, we only
    want "1 month" here).
    """
    @classmethod
    def string_for(cls, value):
        if not isinstance(value, date):
            return value
        if not isinstance(value, datetime):
            value = datetime.combine(value, datetime.min.time())

        # pylint:disable=no-member
        now = datetime.now(timezone.utc if timezone.is_aware(value) else None)
        # Make it as round as we want it:
        relative_delta = relativedelta_rounded(now, value)
        now = value + relative_delta

        if value < now:
            if abs(relative_delta.days) == 1:
                return cls.time_strings["yesterday"]
            delta = now - value
            if delta.days != 0:
                delta_str = timesince(value, now, time_strings=cls.past_substrings)
                return cls.time_strings["past-day"] % {
                    "delta": delta_str.split(", ")[0]
                }
        else:
            if abs(relative_delta.days) == 1:
                return cls.time_strings["tomorrow"]
            delta = value - now
            if delta.days != 0:
                delta_str = timeuntil(value, now, time_strings=cls.future_substrings)
                return cls.time_strings["future-day"] % {
                    "delta": delta_str.split(", ")[0]
                }
        return super().string_for(value)


### TAGS ##################################

@register.tag(name="link")
def do_link(parser: template.base.Parser, token: template.base.Token):
    """
    Outputs an <a> tag. The only purpose of this is to automatically inject
    `target="_blank"` and `rel="external"` if it's an external link, so
    there's a good argument for this being overkill. But I learned stuff from
    writing it, and that's never a waste.

    Has one required positional argument, which is the link URL. This can be a
    raw string or a template variable. Keyword arguments will be used as
    attributes on the <a> element, with underscores replaced by dashes in the
    keys.

    Usage:
    {% link "https://external-link.com" data_foo="bar" %}
    <p>This is my link</p>
    {% endlink %}

    Output:
    <a href="https://external-link.com" target="_blank" data-foo="bar">
        <p>This is my link</p>
    </a>
    """
    bits = token.split_contents()
    if len(bits) < 2:
        raise template.TemplateSyntaxError("'link' template tags must contain a URL.")
    href = parser.compile_filter(bits[1])
    kwargs = token_kwargs(bits[2:], parser)
    nodelist = parser.parse(("endlink",))
    parser.delete_first_token()
    return LinkNode(nodelist, href, **kwargs)


@register.inclusion_tag("klaatu/modals/dynamic.html")
def dynamic_modal(
    modal_id: str,
    url: str = "",
    classes: str = "",
    required_params: str = "",
    optional_params: str = "",
    large=False,
    scrollable=False,
    center=False,
    **kwargs,
):
    """
    Includes the Bootstrap modal "skeleton" from klaatu/modals/dynamic.html
    and renders it in a similar way to `modal()`, except it also sets
    `data-content-url` on the .modal element. On modal show event, the JS
    function loadDynamicContent() will load the contents from this URL into
    the modal.

    The template rendered by this URL will preferably extend
    klaatu/modals/dynamic_content.html.

    Any extra **kwargs, whose key begins with "data_", will be added to the
    .modal element as "data-" attributes.
    For example, `{% dynamic_modal ... data_foo_bar="42" %}` will result in a
    .modal element with `data-foo-bar="42"`.

    Extra kwargs whose keys do _not_ begin with "data_" will be ignored.
    """
    param_list = _get_param_list(required_params, optional_params)
    url = url.strip()
    if url and not url.startswith("/"):
        try:
            url = reverse(url)
        except Exception:
            pass

    return {
        "modal": {
            "id": modal_id.strip(),
            "url": url,
            "classes": classes.strip(),
            "required_params": required_params,
            "optional_params": optional_params,
            "all_params": param_list,
            "large": large,
            "scrollable": scrollable,
            "center": center,
            "data_attrs": {k.replace("_", "-"): v for k, v in kwargs.items() if k.startswith("data_")},
        },
    }


@register.tag(name="section")
def email_section(parser: template.base.Parser, token: template.base.Token):
    """
    Injects the tag's contents into a <table> with one row and one cell,
    roughly mimicking a <div>. This is because email clients are retarded and
    can't handle normal HTML.

    The <table> tag will have class="section"; it is up to the implementation
    to provide such a class. The element is probably best suited for
    graphically "top level" sections, and the class should therefore specify
    "max-width: 600px" or whichever container width you go for. Any kwargs
    will be used for attributes on the lone <td> tag.

    Example:

    {% section class="py-3" style="text-align: center" %}
        <p>Hey ho!</p>
    {% endsection %}

    ... will result in:

    <table class="section">
        <tr>
            <td class="py-3" style="text-align: center">
                <p>Hey ho!</p>
            </td>
        </tr>
    </table>
    """
    bits = token.split_contents()[1:]
    kwargs = {}
    for bit in bits:
        match = template.base.kwarg_re.match(bit)
        if match:
            name, value = match.groups()
            kwargs[name] = parser.compile_filter(value)
    nodelist = parser.parse(("endsection",))
    parser.delete_first_token()
    return EmailSection(nodelist, **kwargs)


@register.simple_tag
def join_query_params(request: HttpRequest, **kwargs) -> str:
    """
    Will coerce the new QueryDict to only contain one value for each key,
    with priority to those in kwargs.
    """
    params = {k: v for k, v in request.GET.dict().items() if not isinstance(v, list)}
    params.update(kwargs)
    querydict = QueryDict(mutable=True)
    querydict.update(params)
    return request.path_info + '?' + querydict.urlencode()


@register.simple_tag(takes_context=True)
def map_to_context(context: template.RequestContext, key: str):
    """
    Tries to get value from current context, both with key as-is and with
    dashes switched to underscores.
    """
    if key in context:
        return context[key]
    if "-" in key:
        return context.get(key.replace("-", "_"), "")
    return ""


@register.simple_tag(takes_context=True)
def modal(
    context: template.RequestContext,
    template_name: str,
    modal_id="",
    classes="",
    required_params="",
    optional_params="",
    footer=True,
    large=False,
    scrollable=False,
    center=False,
    **kwargs
) -> str:
    render_context: dict[str, Any] = {k: v for k, v in context.flatten().items() if isinstance(k, str)}
    render_context.update(kwargs)

    return render_modal(
        template_name=template_name,
        request=context.request,
        modal_id=modal_id,
        classes=classes,
        required_params=required_params,
        optional_params=optional_params,
        footer=footer,
        large=large,
        scrollable=scrollable,
        center=center,
        context=render_context,
    )


@register.inclusion_tag("klaatu/preloader.html")
def preloader(
    id: str | None = None,
    position="absolute",
    show=False,
    large=False,
    framed=True,
    backdrop=True,
    style="",
    **kwargs
):
    """
    Inserts a .preloader element, which will then be shown and hidden by JS.
    Extend klaatu/preloader.html to customize it.

    @param show Show preloader on load
    @param large If False, shrink font & image sizes to 75%; else 100%
    @param framed Whether to include a frame. This may or may not have any
           effect, depending on the HTML/CSS implementation.
    @param backdrop Whether to show the preloader against a backdrop of some
           kind. This may or may not have any effect, depending on the
           HTML/CSS implementation.
    """
    return {
        "preloader": {
            "id": id,
            "show": show,
            "large": large,
            "position": position,
            "framed": framed,
            "backdrop": backdrop,
            "style": style,
            **kwargs,
        }
    }


@register.simple_tag
def static_full_uri(value: str) -> str:
    root_url = getattr(settings, "ROOT_URL", "")
    return urljoin(root_url, static(value))


@register.simple_tag(name='urljoin')
def urljoin_tag(base, url) -> str:
    return urljoin(base, url)


### FILTERS ###############################

@register.filter(name="abs")
def abs_value(value) -> int | None:
    """
    Simply returns the absolute value of `value`, or None if it cannot be
    coerced to integer.
    """
    try:
        return abs(int(value))
    except TypeError:
        return None


@register.filter
def add_str(value, arg) -> str:
    return str(value) + str(arg)


@register.filter
def admin_boolean_icon(value: bool) -> str:
    """
    Taken from django.contrib.admin.templatetags.admin_list._boolean_icon(),
    which is probably considered "internal", so I copied its code instead.
    """
    icon_url = static('admin/img/icon-%s.svg' % {True: 'yes', False: 'no', None: 'unknown'}[value])
    return format_html('<img src="{}" alt="{}">', icon_url, value)


@register.filter(name="capitalize")
@stringfilter
def capitalize_string(value: str) -> str:
    return capitalize(value)


@register.filter
def delta_days(value) -> int | None:
    """
    Return number of days between now and `value`. Positive = future.
    `value` may be a date or datetime object, or a string which can be parsed
    with datetime.fromisoformat().
    """
    if isinstance(value, str):
        try:
            value = datetime.fromisoformat(value)
        except ValueError:
            pass
    if isinstance(value, datetime):
        value = value.date()
    if isinstance(value, date):
        return (value - timezone.localdate()).days
    return None


@register.filter
def distinct(seq: Collection) -> list:
    """Returns unique members of a collection."""
    return list(set(seq))


@register.filter
def divide_by(value, arg):
    try:
        return value / arg
    except Exception:
        return "-"


@register.filter
def emphasize(text: str, words: str | Iterable[str]):
    """
    Make all instances of `words` in `text` <strong>bold</strong>.
    Case insensitive.
    """
    if isinstance(words, str):
        words = words.split(" ")
    pattern = "|".join([w.lower() for w in words])
    return mark_safe(re.sub(f"({pattern})", r"<strong>\1</strong>", text, flags=re.IGNORECASE))


@register.filter
@stringfilter
def full_uri(value: str) -> str:
    root_url = getattr(settings, "ROOT_URL", "")
    return urljoin(root_url, value)


@register.filter
def in_language(value: str, language_code: str) -> str:
    with override(language_code):
        return str(value)


@register.filter
def modulo(value, arg):
    try:
        return int(value) % int(arg)
    except Exception:
        return None


@register.filter
def multiply(value, arg):
    try:
        return str(float(value) * float(arg))
    except Exception:
        return "-"


@register.filter(name="natural_and_list")
def natural_and_list_filter(value: Iterable) -> str:
    return natural_and_list(value)


@register.filter(name="natural_or_list")
def natural_or_list_filter(value: Iterable) -> str:
    return natural_or_list(value)


@register.filter
def naturaltime_short(value):
    return NaturalTimeShortFormatter.string_for(value)


@register.filter
def percent_of(part, whole) -> int:
    """Rounded percentage. Usage: `{{ part|percent_of:whole }}%`"""
    try:
        if not isinstance(part, (int, float)):
            part = int(part)
        if not isinstance(whole, (int, float)):
            whole = int(whole)
        return percent_rounded(part, whole)
    except ValueError:
        return 0


@register.filter
@stringfilter
def render(value: str) -> str:
    try:
        return template.Template(value).render(template.Context())
    except Exception:
        return value


@register.filter
def startswith(value: str, arg: str) -> bool:
    try:
        return value.startswith(arg)
    except Exception:
        return False


@register.filter
def subtract(value, arg):
    try:
        return value - arg
    except Exception:
        return "-"


@register.filter(name="timedelta")
def timedelta_filter(value: Any) -> str:
    if not isinstance(value, (timedelta, float, int)) or not value:
        return "-"
    return timedelta_formatter(value)


@register.filter
def timedelta_rounded(value: Any) -> str:
    if not isinstance(value, (timedelta, float, int)) or not value:
        return "-"
    return timedelta_formatter(value, rounded=True)


@register.filter
def timedelta_rounded_short(value: Any) -> str:
    if not isinstance(value, (timedelta, float, int)) or not value:
        return "-"
    return timedelta_formatter(value, rounded=True, short_format=True)


@register.filter
def timedelta_short(value: Any) -> str:
    if not isinstance(value, (timedelta, float, int)) or not value:
        return "-"
    return timedelta_formatter(value, short_format=True)


@register.filter
def timedelta_time(value: Any) -> str:
    """Returns value in HH:MM:SS format."""
    if not isinstance(value, (timedelta, float, int)):
        return "-"
    if isinstance(value, (int, float)):
        value = timedelta(seconds=value)
    hours = int(value.total_seconds() / 3600)
    minutes = int(value.total_seconds() % 3600 / 60)
    seconds = int(value.total_seconds() % 60)
    return "%02d:%02d:%02d" % (hours, minutes, seconds)



"""
Note: This module does not contain tests, hence its name is 'test' and not
'tests'. It's for extending TestCase.
"""
import json
from typing import Any

from django.core import mail
from django.db.models import Model
from django.http.response import HttpResponse
from django.test import TestCase


class ExtendedTestCase(TestCase):
    def assertGetUrl(self, url: str, expected_url: str | None = None):
        """Assert GET request returns status < 400 and the given URL"""
        response = self.client.get(url, follow=True)
        self.assertLess(response.status_code, 400)
        if expected_url:
            self.assertEqual(response.wsgi_request.path_info, expected_url)
        else:
            self.assertEqual(response.wsgi_request.path_info, url)

    def assertDictContains(self, actual: dict, expected: dict):
        """Assert `actual` contains all keys and values from `expected`"""
        subset = {k: v for k, v in actual.items() if k in expected}
        self.assertDictEqual(subset, expected)

    def assertListContains(self, actual: list, expected: list):
        """Assert `actual` contains all values from `expected`"""
        sublist = [v for v in actual if v in expected]
        self.assertListEqual(sublist, expected)

    def assertDictListContains(self, actual: list, expected: dict):
        """Given a list of dicts, assert it contains `expected`"""
        for d in actual:
            try:
                self.assertDictContains(d, expected)
            except AssertionError:
                pass
            else:
                return
        self.fail(f"{actual} does not contain {expected}")

    def assertModelInstanceContains(self, instance: Model, expected: dict):
        """
        Assert `instance` has the attributes and values from the `expected`
        keys and values, respectively
        """
        for k, v in expected.items():
            try:
                self.assertEqual(getattr(instance, k), v)
            except self.failureException as e:
                raise self.failureException(
                    "instance.%s (%s) != expected[\"%s\"] (%s)" % (k, getattr(instance, k), k, v)
                ) from e

    def assertResponseJSONContains(
        self,
        response: HttpResponse,
        data: Any,
        status_codes: list[int] | None = None,
    ):
        """
        Assert response has any of `status_codes` and has a JSON body that
        either contains `data` (in case `data` is list or dict), or _is_
        `data`.
        """
        content = json.loads(response.content)
        self.assertIn(response.status_code, status_codes or [200, 201])
        if isinstance(content, dict) and isinstance(data, dict):
            self.assertDictContains(content, data)
        elif isinstance(content, list) and isinstance(data, list):
            self.assertListContains(content, data)
        else:
            self.assertEqual(content, data)

    def assertHasReceivedEmail(self, email: str, count: int | None = None):
        """
        Assert that `count` number of emails have been sent to `email`, or
        that _any_ email has been sent to `email` in case `count` is None.
        """
        messages = [m for m in mail.outbox if email in m.to]
        if count is not None:
            self.assertEqual(len(messages), count)
        else:
            self.assertGreaterEqual(len(messages), 1)

    def assertHasNotReceivedEmail(self, email: str):
        self.assertHasReceivedEmail(email, 0)



from typing import Any, Mapping

from django.conf import settings
from django.contrib import messages
from django.forms import Form
from django.forms.utils import ErrorList
from django.http import Http404, HttpRequest, HttpResponse, HttpResponseRedirect
from django.shortcuts import redirect
from django.utils.translation import check_for_language
from django.views.generic.base import View
from django.views.generic.detail import DetailView, SingleObjectMixin
from django.views.generic.edit import FormMixin, FormView, UpdateView


class LanguageMixin:
    """
    To be used with Django REST Framework views. Sets 'language' in the
    context of the serializer (which, preferably, should inherit from
    `klaatu_django.serializer_mixins.LanguageMixin`).
    """
    request: Any

    def get_language(self) -> str:
        """
        Priority:

        1. `language` GET param
        2. Authenticated user's language
        3. `settings.LANGUAGE_CODE_SESSION_KEY` session variable
        4. Default
        """
        assert hasattr(self, 'request')
        LANGUAGE_CODE_SESSION_KEY = getattr(settings, "LANGUAGE_CODE_SESSION_KEY", None)

        if hasattr(self.request, 'query_params'):
            language = self.request.query_params.get('language', None)
            if language and check_for_language(language):
                return language
        if hasattr(self.request, 'user') and hasattr(self.request.user, 'language'):
            return self.request.user.language
        if (
            hasattr(self.request, 'session') and
            LANGUAGE_CODE_SESSION_KEY and
            LANGUAGE_CODE_SESSION_KEY in self.request.session
        ):
            return self.request.session[LANGUAGE_CODE_SESSION_KEY]
        return settings.LANGUAGE_CODE

    def get_serializer_context(self):
        context = super().get_serializer_context()  # type: ignore
        context['language'] = self.get_language()
        return context


class MultipleFormsMixin(FormMixin):
    """
    Inspired by extra_views and https://gist.github.com/michelts/1029336

    initial_data format: {form_prefix: {key: value, ...}, ...}
    form_classes format: {form_prefix: form_class, ...}
    """
    form_classes: Mapping[str, type[Form]] = {}
    request: HttpRequest

    def get_form_classes(self):
        return self.form_classes

    def get_form(self, form_class=None):
        # Otherwise FormMixin.get_form() will mess with us
        return None

    def get_initial_for(self, key: str) -> dict[str, Any]:
        return {}

    def get_form_kwargs_for(self, key: str) -> dict[str, Any]:
        kwargs = {
            "initial": {
                **self.get_initial().get(key, {}),
                **self.get_initial_for(key)
            },
            "prefix": key,
        }
        if self.request.method in ("POST", "PUT"):
            kwargs.update(
                data=self.request.POST,
                files=self.request.FILES,
            )
        return kwargs

    def get_forms(self, form_classes: Mapping[str, type[Form]]) -> dict[str, Form]:
        return {
            key: klass(**self.get_form_kwargs_for(key))
            for key, klass in form_classes.items()
        }

    def get_form_errors(self, forms: dict[str, Form]) -> dict[str, ErrorList]:
        errors: dict[str, ErrorList] = {}
        for prefix, form in forms.items():
            for key, value in form.errors.items():
                if key == "__all__":
                    if "__all__" not in errors:
                        errors["__all__"] = ErrorList()
                    errors["__all__"].extend(value)
                else:
                    errors[f"{prefix}-{key}"] = value
        return errors

    def clean_forms(self, forms: dict[str, Form]) -> bool:
        # TODO: Maybe make it more consistent with Django's clean*() methods
        # (don't return anything but update self.errors instead)
        return all(form.is_valid() for form in forms.values())


class MultipleFormsView(MultipleFormsMixin, FormView):
    """
    Inspired by extra_views and https://gist.github.com/michelts/1029336
    """
    def get(self, request, *args, **kwargs):
        form_classes = self.get_form_classes()
        forms = self.get_forms(form_classes)
        return self.render_to_response(self.get_context_data(forms=forms, **kwargs))

    def post(self, request, *args, **kwargs):
        form_classes = self.get_form_classes()
        forms = self.get_forms(form_classes)
        if self.clean_forms(forms):
            return self.forms_valid(forms)
        return self.forms_invalid(forms)

    def forms_valid(self, forms) -> HttpResponse:
        return HttpResponseRedirect(self.get_success_url())

    def forms_invalid(self, forms) -> HttpResponse:
        return self.render_to_response(self.get_context_data(forms=forms))


class RedirectIfNotFoundMixin(SingleObjectMixin):
    """
    Convenience mixin for redirecting to a chosen URL when an object is not
    found, with an optional error message.
    """
    fallback_to_referer = True
    redirect_message: str | None = None
    redirect_url: str | None = None
    request: HttpRequest

    def get_redirect_message(self) -> str | None:
        return getattr(self, "redirect_message", None)

    def get_redirect_url(self) -> str:
        if self.redirect_url:
            return self.redirect_url
        if self.fallback_to_referer and "referer" in self.request.headers:
            # Ugly hack so we don't redirect back to /login/?next=... or
            # whatever, which in turn would redirect us back here, and so on
            return self.request.headers["referer"].split("?")[0]
        # Last resort:
        raise Http404

    def redirect_with_message(self, exception: Http404 | None = None):
        message = self.get_redirect_message()
        if not message and exception is not None:
            message = str(exception)
        if message:
            messages.error(self.request, message)
        return redirect(self.get_redirect_url())


class BaseRedirectIfNotFoundView(RedirectIfNotFoundMixin, View):
    def dispatch(self, request, *args, **kwargs):
        try:
            return super().dispatch(request, *args, **kwargs)
        except Http404 as e:
            return self.redirect_with_message(e)


class RedirectDetailView(BaseRedirectIfNotFoundView, DetailView):
    pass


class RedirectUpdateView(BaseRedirectIfNotFoundView, UpdateView):
    pass



import logging
import os
import signal
import sys
from pathlib import Path

import django
from django.utils import autoreload

from klaatu_django.settings import gd_settings

DJANGO_AUTORELOAD_ENV = 'RUN_MAIN'
logger = logging.getLogger('django.utils.autoreload')

if gd_settings.RUNSERVER.SERVER == "daphne":
    try:
        from daphne.management.commands.runserver import Command as RunServerCommand  # type: ignore
    except ImportError:
        from django.contrib.staticfiles.management.commands.runserver import Command as RunServerCommand
else:
    from django.contrib.staticfiles.management.commands.runserver import Command as RunServerCommand


class WatchmanReloader(autoreload.WatchmanReloader):
    """
    Don't watch /usr and /etc, mainly since Watchman constantly believes
    /etc/python3.8/sitecustomize.py has been changed for some reason.

    Don't watch anything in a venv package either, for performance reasons.
    """
    def watched_files(self, include_globs=True):
        for f in super().watched_files(include_globs=include_globs):
            if (
                not f.as_posix().startswith('/usr') and
                not f.as_posix().startswith('/etc') and
                not Path(django.__file__).parent.parent in f.parents
            ):
                yield f


class Command(RunServerCommand):  # type: ignore
    default_addr = gd_settings.RUNSERVER.DEFAULT_ADDR
    default_port = gd_settings.RUNSERVER.DEFAULT_PORT

    def run(self, **options):
        if options['use_reloader']:
            run_with_reloader(self.inner_run, **options)
        else:
            self.inner_run(None, **options)


def get_reloader():
    """Return the most suitable reloader for this environment."""
    try:
        WatchmanReloader.check_availability()
    except autoreload.WatchmanUnavailable:
        return autoreload.StatReloader()
    return WatchmanReloader()


def run_with_reloader(main_func, *args, **kwargs):
    signal.signal(signal.SIGTERM, lambda *args: sys.exit(0))
    try:
        if os.environ.get(DJANGO_AUTORELOAD_ENV) == 'true':
            reloader = get_reloader()
            logger.info('Watching for file changes with %s', reloader.__class__.__name__)
            autoreload.start_django(reloader, main_func, *args, **kwargs)
        else:
            exit_code = autoreload.restart_with_reloader()
            sys.exit(exit_code)
    except KeyboardInterrupt:
        pass



import logging

from django.db import models

logger = logging.getLogger(__name__)


class FailSafeManager(models.Manager):
    """Use when successful saving isn't crucial, like for log objects."""
    def create_or_log_error(self, **kwargs):
        try:
            return super().create(**kwargs)
        except Exception:
            logger.error('Could not create %s object', self.model.__name__, exc_info=True)
            return None

    def create_quietly(self, **kwargs):
        try:
            return super().create(**kwargs)
        except Exception:
            return None



import io
import logging
import pickle

from PIL import Image

from django.core.exceptions import ValidationError
from django.core.files import File
from django.db import models
from django.db.models import BinaryField, Case, F, FloatField, IntegerField, Q, Value as V, When
from django.db.models.expressions import Combinable, Expression
from django.db.models.fields.files import ImageFieldFile
from django.db.models.functions import Cast, Round

logger = logging.getLogger(__name__)


def TrueIf(*args, **kwargs) -> Expression:
    """Convenience method: 'if all the arguments are true, then true'."""
    return Case(When(Q(*args, **kwargs), then=V(True)), default=V(False))


def TrueIfAny(*args, **kwargs) -> Expression:
    """If _any_ of the arguments is true, then true."""
    conds = Q()
    for arg in args:
        conds |= Q(arg)
    for key, value in kwargs.items():
        conds |= Q(**{key: value})
    return Case(When(conds, then=V(True)), default=V(False))


def CorrectRound(field: str | Combinable) -> Expression:
    """
    Correctly rounds a numeric field to int.

    Formula: int(value + (int(value) % 2))
    So:
    CorrectRound(4.5) = int(4.5 + (4 % 2)) = int(4.5 + 0) = 4
    CorrectRound(3.5) = int(3.5 + (3 % 2)) = int(3.5 + 1) = 4
    """
    if isinstance(field, str):
        field = F(field)
    return Cast(
        field + (Cast(field, IntegerField()) % V(2)),
        # field + Mod(Cast(field, IntegerField()), V(2)),
        IntegerField()
    )


def PercentRounded(part: str, whole: str) -> Expression:
    """
    Given two numeric SQL fields, return a percentage as integer.
    """
    return Case(
        When(Q(**{whole: 0}), then=V(0)),
        default=Round(Cast(part, FloatField()) / F(whole) * 100, output_field=IntegerField())
    )


class ResizeImageFieldFile(ImageFieldFile):
    """Resize large images, silently report error on fail"""
    max_width: int | None
    max_height: int | None

    def __init__(self, instance, field, name):
        assert isinstance(field, ResizeImageField)
        self.max_height, self.max_width = field.max_height, field.max_width
        super().__init__(instance, field, name)

    def get_target_size(self, image: Image.Image) -> tuple[int, int]:
        width_divider = image.width / (self.max_width or image.width)
        height_divider = image.height / (self.max_height or image.height)
        divider = max([width_divider, height_divider])
        return int(image.width / divider), int(image.height / divider)

    def should_resize(self, image: Image.Image) -> bool:
        return (
            (self.max_width is not None and image.width > self.max_width) or
            (self.max_height is not None and image.height > self.max_height)
        )

    def save(self, name: str, content, save=True):
        super().save(name, content, save=save)
        content_image = getattr(content, "image", None)
        if content_image:
            if self.should_resize(content_image):
                try:
                    image = Image.open(self.file)
                    fp = io.BytesIO()
                    resized = image.resize(self.get_target_size(content_image))
                    resized.save(fp, format=image.format)
                    if self.name:
                        self.storage.delete(self.name)
                    self.save(name, File(fp), save=save)
                    fp.close()
                except Exception:
                    logger.error(
                        "ResizeImageFieldFile: Could not resize image",
                        exc_info=True,
                        extra={"filename": name, "instance": self.instance}
                    )


class ResizeImageField(models.ImageField):
    attr_class = ResizeImageFieldFile

    def __init__(self, max_height=None, max_width=None, **kwargs):
        self.max_height, self.max_width = max_height, max_width
        super().__init__(**kwargs)


class TruncatedCharField(models.CharField):
    """Use for char fields that aren't super important, like in logs."""
    def to_python(self, value):
        value = super().to_python(value)
        if value and self.max_length and len(value) > self.max_length:
            logger.warning(
                "Value of TruncatedCharField '%s' exceeds max_length %d (%d)",
                self.name,
                self.max_length,
                len(value),
                extra={"model": getattr(self, "model", None), "value": value}
            )
            return value[:self.max_length]
        return value


class TruncatedURLField(TruncatedCharField, models.URLField):
    ...


class PickleField(BinaryField):
    description = "Auto-pickled and unpickled object"

    def get_prep_value(self, value):
        return pickle.dumps(value)

    def from_db_value(self, value, expression, connection):
        if value is None:
            return value
        return pickle.loads(value)

    def to_python(self, value):
        if isinstance(value, bytes):
            try:
                return pickle.loads(value)
            except Exception as e:
                raise ValidationError(f"{e.__class__} error in unpickling: {str(e)}") from e
        return value



import inspect
from io import TextIOWrapper

from django.core.files import locks
from django.core.management.base import BaseCommand as DjangoBaseCommand, CommandParser


class BaseCommand(DjangoBaseCommand):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        setattr(self, "_handle", self.handle)
        setattr(self, "handle", self.real_handle)

    def add_arguments(self, parser: CommandParser):
        super().add_arguments(parser)
        parser.add_argument(
            "--ipdb",
            action="store_true",
            help="Drop into IPDB shell at the start of execution.",
        )

    def real_handle(self, *args, **options):
        if options["ipdb"]:
            try:
                import ipdb
                ipdb.runcall(self._handle, *args, **options)
            except ImportError as e:
                raise Exception("--ipdb flag set, but ipdb could not be imported. Exiting!") from e
        else:
            self._handle(*args, **options)

    def _handle(self, *args, **options):
        raise NotImplementedError("Default _handle() should have been overwritten by now?!")


class ExclusiveCommand(BaseCommand):
    """
    For management commands that should be blocked from concurrent execution.
    Uses a file lock as an exclusivity check, by default on the file
    containing the command class (which is really the most logical choice).

    Thanks to some ugl^H^H^Hbeautiful monkey patching, child classes don't
    need to implement any new methods or really change anything; just inherit
    this class instead of BaseCommand, and forget about it.
    """
    lockfile: str
    _lockfile: TextIOWrapper

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        setattr(self, "_handle", self.handle)
        setattr(self, "handle", self.real_handle)

    def real_handle(self, *args, **options):
        if hasattr(self, "lockfile"):
            lockfile = self.lockfile
        else:
            lockfile = inspect.getfile(self.__class__)
        with open(lockfile) as self._lockfile:
            if locks.lock(self._lockfile, locks.LOCK_EX | locks.LOCK_NB):
                try:
                    self._handle(*args, **options)
                finally:
                    locks.unlock(self._lockfile)
            else:
                self.stderr.write("This command is already running in another process.")

    def _handle(self, *args, **options):
        raise NotImplementedError("Default _handle() should have been overwritten by now?!")



from typing import TYPE_CHECKING, Sequence

from django.contrib import admin
from django.contrib.admin.options import InlineModelAdmin
from django.contrib.admin.sites import AdminSite
from django.contrib.admin.utils import quote
from django.db.models import Model, QuerySet
from django.db.models.fields.reverse_related import ForeignObjectRel
from django.http import HttpRequest
from django.urls import reverse
from django.utils.html import format_html

from .typing import AdminFieldsetsType, AdminFieldsType, FormType

if TYPE_CHECKING:
    from django.utils.functional import _StrPromise  # type: ignore


class BooleanListFilter(admin.SimpleListFilter):
    def lookups(self, request, model_admin):
        return [('1', 'Yes'), ('0', 'No')]

    def queryset(self, request, queryset):
        value = self.value()
        if value == '1':
            return queryset.filter(**{self.parameter_name: True})
        if value == '0':
            return queryset.filter(**{self.parameter_name: False})
        return queryset


class NoDeleteActionMixin:
    def get_actions(self, request):
        actions = super().get_actions(request)  # type: ignore
        if 'delete_selected' in actions:
            del actions['delete_selected']
        return actions


class SetCreatedByAdmin(admin.ModelAdmin):
    # For use in admin pages for models with `created_by` fields
    def save_model(self, request, obj, form, change):
        if not change:
            try:
                obj.created_by = request.user
            except AttributeError:
                pass
        super().save_model(request, obj, form, change)


class SetCreatedByInlineAdmin(admin.ModelAdmin):
    # Set `created_by` on inline objects in admin
    def save_formset(self, request, form, formset, change):
        formset.save(commit=False)
        for obj in formset.new_objects:
            try:
                obj.created_by = request.user
            except AttributeError:
                pass
        super().save_formset(request, form, formset, change)


class TabularManyToManyInline(admin.TabularInline):
    template = "admin/edit_inline/tabular_manytomany.html"


class RelatedLinkMixin:
    admin_site: AdminSite

    def get_related_changeform_link(self, related_obj: Model | None, display_attr: str | None = None):
        """
        If `related_obj` is a model instance, and its model is registered with
        the AdminSite, return a link to the changeform for this instance.
        Otherwise, just return `related_obj` as-is.

        If `display_attr` is set, `related_obj` is displayed using this
        attribute on the model. Otherwise by the model's __str__().

        Usage:

        list_display = (..., "user_link", ...)

        @admin.display(description="user", ordering="user__username")
        def user_link(self, obj):
            return self.get_related_changeform_link(obj.user, "username")
        """
        if (
            isinstance(related_obj, Model) and
            self.admin_site.is_registered(related_obj._meta.model) and
            related_obj._meta.pk
        ):
            pk = getattr(related_obj, related_obj._meta.pk.attname)
            return format_html(
                '<a href="{}">{}</a>',
                reverse(
                    "admin:%s_%s_change" % (related_obj._meta.app_label, related_obj._meta.model_name),
                    args=(quote(pk),),
                    current_app=self.admin_site.name
                ),
                getattr(related_obj, display_attr) if display_attr else related_obj
            )
        return related_obj

    def get_related_changeform_link_list(
        self,
        related_queryset: QuerySet,
        display_attr: str | None = None,
        container_classes: str = "",
        item_classes: str = "",
    ):
        """
        Returns links to changeform for each item in `related_queryset`,
        each in their own <div class="related-changeform-link-list-item">. The
        whole list is enclosed in a <div class="related-changeform-link-list">.
        If queryset's model is not registered with the AdminSite, returns the
        same list but without links.

        If `display_attr` is set, items are displayed using this attribute
        on the model instances. Otherwise by the model's __str__().

        `container_classes` sets extra CSS classes on the containing <div>
        element, `item_classes` sets them on each item's <div>.

        Usage:

        list_display = (..., "user_link_list", ...)

        @admin.display(description="users", ordering="users__username")
        def user_link_list(self, obj):
            return self.get_related_changeform_link_list(
                obj.users.all(),
                "username"
            )
        """
        if self.admin_site.is_registered(related_queryset.model):
            opts = related_queryset.model._meta
            html = "".join([
                '<div class="related-changeform-link-list-item %s"><a href="%s">%s</a></div>' % (
                    item_classes,
                    reverse(
                        "admin:%s_%s_change" % (opts.app_label, opts.model_name),
                        args=(quote(getattr(obj, opts.pk.attname)),),
                        current_app=self.admin_site.name
                    ),
                    getattr(obj, display_attr) if display_attr else obj
                )
                for obj in related_queryset
            ])
        else:
            html = "".join([
                f'<div class="related-changeform-link-list-item {item_classes}">' +
                str(getattr(obj, display_attr)) if display_attr else str(obj) +
                "</div>"
                for obj in related_queryset
            ])

        return format_html(f'<div class="related-changeform-link-list {container_classes}">{html}</div>')

    def get_related_changelist_link(
        self,
        obj: Model | None,
        related_name: str,
        proxy_model: type[Model] | None = None,
        show_zero: bool = True,
        verbose_name: "_StrPromise | str | None" = None,
        verbose_name_plural: "_StrPromise | str | None" = None
    ):
        """
        `related_name` should be the name of a many-to-many or many-to-one
        (reverse foreign key relationship) field on `obj`. If the related
        model is registered with the AdminSite, return a link to that model's
        changelist, filtered for those instances that are related to `obj`.

        Set `proxy_model` if the link should go to the admin for a proxy
        model, rather than the concrete model indicated by the relation.

        Link text will be "[number of instances] [verbose_name[_plural]]",
        where `verbose_name[_plural]` is either set in the method arguments or
        collected from the related model's meta attributes. The usage of the
        "_plural" suffix is of course dependent upon the number of related
        objects found.

        Usage:

        list_display = (..., "user_list_link", ...)

        @admin.display(description="users")
        def user_list_link(self, obj):
            return self.get_related_changelist_link(obj, "users")
        """
        if obj is None or obj._meta.pk is None:
            return None

        related_field = obj._meta.get_field(related_name)

        if isinstance(related_field, ForeignObjectRel):
            if proxy_model and related_field.related_model and issubclass(proxy_model, related_field.related_model):
                related_model = proxy_model
            else:
                related_model = related_field.related_model

            if related_model and self.admin_site.is_registered(related_model):
                opts = related_model._meta
                verbose_name = verbose_name or related_model._meta.verbose_name
                verbose_name_plural = verbose_name_plural or related_model._meta.verbose_name_plural
                obj_count = getattr(obj, related_name).count()

                if not obj_count and not show_zero:
                    return ""

                return format_html(
                    '<a href="{}?{}__exact={}">{} {}</a>',
                    reverse(
                        "admin:%s_%s_changelist" % (opts.app_label, opts.model_name),
                        current_app=self.admin_site.name
                    ),
                    related_field.field.get_attname(),
                    getattr(obj, obj._meta.pk.attname),
                    obj_count,
                    verbose_name if obj_count == 1 else verbose_name_plural
                )

        return None


class ExtendedModelAdmin(RelatedLinkMixin, admin.ModelAdmin):
    """Meant to collect all kinds of useful extra functionality."""


class ExtendedTabularInline(RelatedLinkMixin, admin.TabularInline):
    pass


class ExtendedStackedInline(RelatedLinkMixin, admin.StackedInline):
    pass


class SeparateAddMixin(admin.ModelAdmin):
    """
    Use custom settings when a change form is used for adding a new object,
    rather than editing an existing one. Mirrors the `exclude`, `fields`,
    `fieldsets`, `form`, `inlines`, and `readonly_fields` properties. If a
    property is None, the regular one will be used instead.
    """
    add_exclude: Sequence[str] | None = None
    add_fields: AdminFieldsType | None = None
    add_fieldsets: AdminFieldsetsType | None = None
    add_form: FormType | None = None
    add_inlines: Sequence[type[InlineModelAdmin]] | None = None
    add_readonly_fields: Sequence[str] | None = None

    def get_add_exclude(self, request: HttpRequest) -> Sequence[str] | None:
        return self.add_exclude

    def get_add_fields(self, request: HttpRequest) -> AdminFieldsType | None:
        return self.add_fields

    def get_add_fieldsets(self, request: HttpRequest) -> AdminFieldsetsType | None:
        return self.add_fieldsets

    def get_add_inlines(self, request: HttpRequest) -> Sequence[type[InlineModelAdmin]] | None:
        return self.add_inlines

    def get_add_readonly_fields(self, request: HttpRequest) -> Sequence[str] | None:
        return self.add_readonly_fields

    def get_exclude(self, request, obj=None):
        if obj is None:
            add_exclude = self.get_add_exclude(request)
            if add_exclude is not None:
                return add_exclude
        return super().get_exclude(request, obj)

    def get_fields(self, request, obj=None):
        if obj is None:
            add_fields = self.get_add_fields(request)
            if add_fields is not None:
                return add_fields
        return super().get_fields(request, obj)

    def get_fieldsets(self, request, obj=None):
        if obj is None:
            add_fieldsets = self.get_add_fieldsets(request)
            if add_fieldsets is not None:
                return add_fieldsets
        return super().get_fieldsets(request, obj)

    def get_form(self, request, obj=None, change=False, **kwargs):
        defaults = {}
        if self.add_form is not None and not change:
            defaults['form'] = self.add_form
        defaults.update(kwargs)
        return super().get_form(request, obj, change=change, **defaults)

    def get_inlines(self, request, obj):
        if obj is None:
            add_inlines = self.get_add_inlines(request)
            if add_inlines is not None:
                return add_inlines
        return super().get_inlines(request, obj)  # type: ignore

    def get_readonly_fields(self, request, obj=None):
        if obj is None:
            add_readonly_fields = self.get_add_readonly_fields(request)
            if add_readonly_fields is not None:
                return add_readonly_fields
        return super().get_readonly_fields(request, obj)



from django.core.management.base import CommandParser
from django.utils.module_loading import import_string

from klaatu_django.management.base import BaseCommand


class Command(BaseCommand):
    classes: dict[type, list[type]] = {}
    help = 'Outputs a reverse inheritance tree for a class.'

    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument('class_path', help='Full path (period-separated) of a class.')

    def handle(self, *args, **options):
        start = import_string(options['class_path'])
        self.get_bases(start)
        self.print_class(start)

    def get_bases(self, klass: type):
        if klass not in self.classes:
            bases = [b for b in klass.__bases__ if b is not object]
            self.classes[klass] = bases
            for base in bases:
                self.get_bases(base)

    def print_class(self, klass: type, depth=0, prefix='', is_last_child=False):
        if depth:
            self.stdout.write(prefix, ending='')
            if is_last_child:
                self.stdout.write('└── ', ending='')
            else:
                self.stdout.write('├── ', ending='')

        self.stdout.write(f'{klass.__module__}.{klass.__name__}')

        if not depth:
            child_prefix = ''
        elif is_last_child:
            child_prefix = prefix + '    '
        else:
            child_prefix = prefix + '│   '

        children = self.classes.get(klass, [])

        for idx, child in enumerate(children):
            self.print_class(child, depth + 1, child_prefix, idx == len(children) - 1)



from rest_framework.routers import DynamicRoute, Route, SimpleRouter


class SingletonRouter(SimpleRouter):
    """
    Router where the "root" route (without lookup args in the URL) is actually
    a detail view. It's up to the viewset to implement get_object() in a way
    that does not require any PK kwargs.
    """
    routes = [
        Route(
            url=r'^{prefix}{trailing_slash}$',
            mapping={
                'get': 'retrieve',
                'post': 'create',
                'put': 'update',
                'patch': 'partial_update',
                'delete': 'destroy',
            },
            name='{basename}-detail',
            detail=True,
            initkwargs={'suffix': 'Instance'}
        ),
        DynamicRoute(
            url=r'^{prefix}/{url_path}{trailing_slash}$',
            name='{basename}-{url_name}',
            detail=True,
            initkwargs={}
        ),
    ]



from rest_framework.authentication import BasicAuthentication
from rest_framework.exceptions import AuthenticationFailed


class AdminBasicAuthentication(BasicAuthentication):
    """Extending the basic auth to only allow superuser accounts"""
    def authenticate_credentials(self, userid, password, request=None):
        user, _ = super().authenticate_credentials(userid, password, request=request)
        if not getattr(user, "is_superuser", False):
            raise AuthenticationFailed('User is not superuser.')
        return (user, None)



from typing import Sequence, TypeVar

from django.forms import Form

AdminFieldsType = Sequence[str | Sequence[str]]

AdminFieldsetsType = Sequence[tuple[str | None, dict[str, str | AdminFieldsType]]]

_Form = TypeVar("_Form", bound=Form)

FormType = type[_Form]



import functools
from typing import Any

from rest_framework import serializers

from django.conf import settings
from django.contrib.auth import get_user_model
from django.utils.translation import override

from .serializer_fields import FileURLField

User = get_user_model()


class UserMixin:
    def get_user(self):
        context = getattr(self, 'context', {})
        assert 'request' in context, 'request must be included in serializer context'
        assert isinstance(context['request'].user, User), \
            f'context["request"].user must be a {User.__module__}.{User.__name__} object'
        return context['request'].user


class CreatedByMixin:
    """
    Sets `created_by` on a created object to the authenticated user. Of course
    only makes sense if there is such a field on the model.
    """
    def create(self, validated_data):
        validated_data = validated_data or {}
        context = getattr(self, 'context', {})
        try:
            user = context['request'].user
            if isinstance(user, User):
                validated_data.update(created_by=user)
        except (AttributeError, KeyError):
            pass
        return super().create(validated_data)  # type: ignore


def override_language(func):
    """
    Used by LanguageMixin below, but could of course be used for any
    serializer method that needs it.
    """
    @functools.wraps(func)
    def wrapper(serializer, *args, **kwargs):
        with override(serializer.context.get('language', settings.LANGUAGE_CODE)):
            return func(serializer, *args, **kwargs)
    return wrapper


class LanguageMixin:
    """
    Set self.context['language'] to a language code in order to present the
    response in that language.

    You probably want the corresponding view to inherit from
    `klaatu_django.view_mixins.LanguageMixin`.
    """
    context: dict[str, Any]

    @override_language
    def to_representation(self, instance):
        return super().to_representation(instance)  # type: ignore


class ImageSerializer(serializers.Serializer):
    """
    Not really a mixin, but placed here for silly circular import reasons.

    Requires the parent object to have a file field called 'image', and a
    char field called 'image_alt'.
    """
    url = FileURLField(source='image.url')
    alt = serializers.CharField(source='image_alt')

    def to_representation(self, instance):
        if not instance.image:
            return None
        return super().to_representation(instance)



from urllib.parse import urljoin

from rest_framework import serializers

from django.conf import settings


class FileURLField(serializers.URLField):
    def get_attribute(self, instance):
        try:
            return urljoin(settings.ROOT_URL, super().get_attribute(instance))  # type: ignore
        except ValueError:
            return None



import copy
import os
import re
import time
from abc import ABCMeta
from datetime import datetime, timedelta
from importlib import import_module
from os.path import basename, splitext
from statistics import mean, median
from types import ModuleType
from typing import TYPE_CHECKING, Any, Iterable, TypeVar

from bs4 import BeautifulSoup
from dateutil.relativedelta import relativedelta
from rest_framework.serializers import SerializerMetaclass

from django.conf import settings
from django.core.exceptions import ValidationError, ViewDoesNotExist
from django.core.serializers.json import DjangoJSONEncoder
from django.core.validators import validate_email
from django.db.models import DurationField, Model, QuerySet
from django.db.models.functions import Cast
from django.http import HttpRequest
from django.template.loader import render_to_string
from django.urls import URLPattern, URLResolver
from django.utils import timezone
from django.utils.safestring import mark_safe
from django.utils.translation import get_language, gettext, ngettext

if TYPE_CHECKING:
    from django.utils.functional import _StrPromise  # type: ignore

_T = TypeVar("_T")


class CastToDuration(Cast):
    VALID_UNITS = [
        {'name': 'microsecond', 'plural': 'microseconds', 'multiplier': 1},
        {'name': 'millisecond', 'plural': 'milliseconds', 'multiplier': 1000},
        {'name': 'second', 'plural': 'seconds', 'multiplier': 1_000_000},
        {'name': 'minute', 'plural': 'minutes', 'multiplier': 60_000_000},
        {'name': 'hour', 'plural': 'hours', 'multiplier': 3_600_000_000},
        {'name': 'day', 'plural': 'days', 'multiplier': 3_600_000_000 * 24},
        {'name': 'week', 'plural': 'weeks', 'multiplier': 3_600_000_000 * 24 * 7},
        {'name': 'month', 'plural': 'months', 'multiplier': 3_600_000_000 * 24 * 30},
        {'name': 'year', 'plural': 'years', 'multiplier': 3_600_000_000 * 24 * 365},
        {'name': 'decade', 'plural': 'decades', 'multiplier': 3_600_000_000 * 24 * 365 * 10},
        {'name': 'century', 'plural': 'centuries', 'multiplier': 3_600_000_000 * 24 * 365 * 100},
        {'name': 'millennium', 'plural': 'millennia', 'multiplier': 3_600_000_000 * 24 * 365 * 1000},
    ]

    def __init__(self, expression, unit: str):
        for valid_unit in self.VALID_UNITS:
            if unit in (valid_unit['name'], valid_unit['plural']):
                self.unit = valid_unit
                break
        if not hasattr(self, 'unit'):
            raise ValueError(f'"{unit}" is not a correct unit for CastToDuration.')
        super().__init__(expression, DurationField())

    def as_postgresql(self, compiler, connection, **extra_context):
        extra_context.update(unit=self.unit['name'])
        return self.as_sql(  # type: ignore
            compiler,
            connection,
            template='(%(expressions)s || \' %(unit)s\')::%(db_type)s',
            **extra_context
        )

    def as_sqlite(self, compiler, connection, **extra_context):
        extra_context.update(multiplier=self.unit['multiplier'])
        template = '%(function)s(%(expressions)s * %(multiplier)d AS %(db_type)s)'
        return super().as_sqlite(compiler, connection, template=template, **extra_context)


class FailSafeJSONEncoder(DjangoJSONEncoder):
    """
    A JSON encoder that will not fail, but its results cannot be used to
    reliably restore the original data.
    """
    def default(self, o):
        try:
            return super().default(o)
        except TypeError as e:
            return str(e)


class Lock:
    """Does the same as `lock`, but as a context manager."""
    def __init__(self, lockfile: str):
        self.lockfile = lockfile

    def __enter__(self):
        if os.path.exists(self.lockfile):
            raise LockException(f"Could not acquire lockfile: {self.lockfile}")
        with open(self.lockfile, "w") as f:
            f.write("LOCKED")

    def __exit__(self, *args, **kwargs):
        remote_attempts = 0
        while os.path.exists(self.lockfile) and remote_attempts < 10:
            os.remove(self.lockfile)
            remote_attempts += 1


class LockException(Exception):
    ...


class ObjectJSONEncoder(DjangoJSONEncoder):
    """
    Somewhat enhanced JSON encoder, for when you want that sort of thing.
    Represents Django models with their string representations and binary
    values are redacted, so it cannot be used to reliably load a previous
    dump.
    """
    def default(self, o):
        if isinstance(o, Model):
            return str(o.pk)
        if isinstance(o, QuerySet):
            return list(o)
        if isinstance(o, bytes):
            return "[Binary data]"
        if isinstance(o, set):
            o = list(o)
        try:
            return super().default(o)
        except TypeError as ex:
            if hasattr(o, '__dict__'):
                return o.__dict__
            raise ex


class SerializerABCMeta(SerializerMetaclass, ABCMeta):
    """
    To be used with "abstract" base serializer classes. Usage:

    from abc import ABC
    from rest_framework.serializers import Serializer

    class MyBaseSerializer(Serializer, ABC, metaclass=SerializerABCMeta):
        ...
    """


def capitalize(string: "str | _StrPromise | None", language: str | None = None) -> str:
    """
    Language-dependent word capitalization. For English, it capitalizes every
    word except some hard-coded exceptions (the first and last word are always
    capitalized, however). For all other languages, only the first word.

    Side effect: Will replace multiple consecutive spaces with only one space.

    @param language Optional, will use current session's language if not set.
    """
    language = language or get_language()

    if string is None:
        return ""

    if language == "en":
        non_capped = ['a', 'an', 'and', 'but', 'for', 'from', 'if', 'nor', 'of', 'or', 'so', 'the']
        words = string.split(" ")
        for idx, word in enumerate(words):
            if word and (idx == 0 or idx == len(words) - 1 or re.sub(r"\W", "", word).lower() not in non_capped):
                words[idx] = word[0].upper() + word[1:]
        return " ".join(words)
    return string.capitalize()


def extract_views_from_urlpatterns(
    urlpatterns=None,
    base="",
    namespace=None,
    app_name=None,
    app_names=None,
    only_parameterless=False,
    urlkwargs=None,
):
    views = {}
    if urlpatterns is None:
        root_urlconf = import_module(settings.ROOT_URLCONF)
        assert hasattr(root_urlconf, "urlpatterns")
        urlpatterns = getattr(root_urlconf, "urlpatterns", [])
        app_name = root_urlconf.__package__
    for p in urlpatterns:
        if isinstance(p, URLPattern) and (app_names is None or app_name in app_names):
            try:
                if only_parameterless and p.pattern.regex.groups > 0:
                    continue
                if p.name and namespace:
                    view_name = f"{namespace}:{p.name}"
                elif p.name:
                    view_name = p.name
                else:
                    continue
                views[view_name] = {
                    "app_name": app_name,
                    "url": base + str(p.pattern),
                    "urlkwargs": (urlkwargs or []) + list(p.pattern.regex.groupindex),
                }
            except ViewDoesNotExist:
                continue
        elif isinstance(p, URLResolver):
            # Hack: Never include admin urls
            if p.app_name == "admin" or (only_parameterless and p.pattern.regex.groups > 0):
                continue
            try:
                patterns = p.url_patterns
            except ImportError:
                continue
            if namespace and p.namespace:
                _namespace = f"{namespace}:{p.namespace}"
            else:
                _namespace = p.namespace or namespace
            if isinstance(p.urlconf_module, ModuleType):
                try:
                    _app_name = p.urlconf_module.app_name
                except AttributeError:
                    _app_name = p.urlconf_module.__package__
            else:
                _app_name = app_name
            views.update(
                extract_views_from_urlpatterns(
                    urlpatterns=patterns,
                    base=base + str(p.pattern),
                    namespace=_namespace,
                    app_name=_app_name,
                    app_names=app_names,
                    only_parameterless=only_parameterless,
                    urlkwargs=(urlkwargs or []) + list(p.pattern.regex.groupindex),
                )
            )
    return dict(sorted(views.items(), key=lambda kv: kv[1]['app_name'] + kv[0]))


def get_client_ip(meta_dict: dict[str, Any]) -> str | None:
    """
    Very basic, but still arguably does a better job than `django-ipware`, as
    that one doesn't take port numbers into account.

    For use with HttpRequest, send `request.META`.
    """
    meta_keys = (
        'HTTP_X_FORWARDED_FOR',
        'X_FORWARDED_FOR',
        'HTTP_CLIENT_IP',
        'HTTP_X_REAL_IP',
        'HTTP_X_FORWARDED',
        'HTTP_X_CLUSTER_CLIENT_IP',
        'HTTP_FORWARDED_FOR',
        'HTTP_FORWARDED',
        'HTTP_VIA',
        'REMOTE_ADDR',
    )
    value = None
    for key in meta_keys:
        meta_value = meta_dict.get(key, None)
        if isinstance(meta_value, str):
            value = meta_value.split(':')[0]
            if value:
                break
    return value


def is_url_name(value: str) -> bool:
    """
    Really just a guess, based on a somewhat consistent naming of Django URLs.
    """
    return bool(re.match(r"^[a-zA-Z0-9_\-:]+$", value))


def is_valid_email(value: Any) -> bool:
    try:
        validate_email(value)
    except (ValidationError, TypeError):
        return False
    return True


def natural_and_list(items: Iterable, enclose_items_in_tag="") -> str:
    return natural_list(items, enclose_items_in_tag=enclose_items_in_tag)


def natural_list(items: Iterable, or_separated=False, enclose_items_in_tag="") -> str:
    """
    Turns `items` into a natural-language string. Will be "or"-separated if
    `or_separated == True`, else "and"-separated. Of course, the English
    original strings use the Oxford comma.

    Example:
    In [1]: natural_list(
       ...:     ["foo", "bar", "baz"],
       ...:     or_separated=True,
       ...:     enclose_items_in_tag="em"
       ...: )
    Out[1]: '<em>foo</em>, <em>bar</em>, or <em>baz</em>'
    """
    def enclose(item):
        if enclose_items_in_tag:
            return "<%s>%s</%s>" % (enclose_items_in_tag, item, enclose_items_in_tag)
        return str(item)

    item_list = list(items)
    if len(item_list) == 0:
        return ""
    if len(item_list) == 1:
        return enclose(item_list[0])
    if len(item_list) == 2:
        return "%s %s %s" % (
            enclose(item_list[0]),
            gettext("or") if or_separated else gettext("and"),
            enclose(item_list[1]),
        )
    vars = {"list": ", ".join([enclose(i) for i in item_list[:-1]]), "last_item": enclose(item_list[-1])}
    if or_separated:
        # Translators: %(list)s is a comma-separated list of 2 or more items.
        return gettext("%(list)s, or %(last_item)s") % vars
    # Translators: %(list)s is a comma-separated list of 2 or more items.
    return gettext("%(list)s, and %(last_item)s") % vars


def natural_or_list(items: Iterable, enclose_items_in_tag="") -> str:
    return natural_list(items, or_separated=True, enclose_items_in_tag=enclose_items_in_tag)


def relativedelta_rounded(dt1: datetime, dt2: datetime) -> relativedelta:
    """
    Rounds to the nearest "time unit", using perhaps arbitrary algorithms.
    """
    # First make sure both are naive OR aware:
    if timezone.is_naive(dt1) and not timezone.is_naive(dt2):
        dt1 = timezone.make_aware(dt1)
    elif timezone.is_naive(dt2) and not timezone.is_naive(dt1):
        dt2 = timezone.make_aware(dt2)
    delta = relativedelta(dt1, dt2)
    # >= 1 months or >= 25 days: return years + rounded months
    if delta.years or delta.months or delta.days >= 25:
        return relativedelta(years=delta.years, months=delta.months + round(delta.days / 30))
    # 7 - 24 days: return rounded weeks
    if delta.days >= 7:
        return relativedelta(weeks=round(delta.days / 7))
    # Dates are different: return that difference as number of days
    if dt1.day != dt2.day:
        return relativedelta(
            datetime(dt1.year, dt1.month, dt1.day),
            datetime(dt2.year, dt2.month, dt2.day)
        )
    # >= 1 hour: return rounded hours
    if delta.hours:
        return relativedelta(hours=delta.hours + round(delta.minutes / 60))
    # >= 1 minute: return minutes (not rounded!)
    if delta.minutes:
        return relativedelta(minutes=delta.minutes)
    # Don't bother with microseconds :P
    return delta


def render_modal(
    template_name: str,
    request: HttpRequest | None = None,
    modal_id="",
    classes="",
    required_params="",
    optional_params="",
    footer=True,
    large=False,
    scrollable=False,
    center=False,
    context: dict[str, Any] | None = None,
):
    """
    Gets a Bootstrap modal from the template file `template_name`, renders it
    with context from the parameters, and returns the result. The template
    file will preferably extend klaatu/modals/base.html.

    `required_params` and `optional_params` are there to tell the JS function
    openModalOnLoad() which GET params to look for. The required ones will
    be injected as `data-required-params` on the .modal element in base.html,
    the optional ones as `data-optional-params`. All those parameters will be
    stripped from the URL when openModalOnLoad() is finished. The only
    difference between the two kinds is that without the required parameters,
    openModalOnLoad() will refuse to open the modal.
    """
    param_list = _get_param_list(required_params, optional_params)

    if not modal_id:
        modal_id = splitext(basename(template_name))[0].replace("_", "-") + "-modal"

    context = context or {}
    context["modal"] = {
        "required_params": required_params,
        "optional_params": optional_params,
        "all_params": param_list,
        "id": modal_id,
        "classes": classes,
        "footer": footer,
        "large": large,
        "scrollable": scrollable,
        "center": center,
    }
    return mark_safe(render_to_string(template_name=template_name, context=context, request=request))


def simple_pformat(obj: Any, indent: int = 4, current_depth: int = 0) -> str:
    """
    Pretty formatter that outputs stuff the way I want it, no more, no less
    """
    def format_value(v: Any) -> str:
        return f"'{v}'" if isinstance(v, str) else repr(v)

    def is_scalar(v: Any) -> bool:
        return isinstance(v, str) or not isinstance(v, Iterable)

    def format_dict(obj: dict) -> str:
        ret = "{"
        if len(obj) > 0:
            multiline = len(obj) > 1 or (len(obj) == 1 and not is_scalar(list(obj.values())[0]))
            if multiline:
                ret += "\n"
            for key, value in obj.items():
                if multiline:
                    ret += " " * (current_depth * indent + indent)
                ret += format_value(key) + ": "
                ret += simple_pformat(value, indent=indent, current_depth=current_depth + 1)
                if multiline:
                    ret += ",\n"
            if multiline:
                ret += " " * (current_depth * indent)
        ret += "}"
        return ret

    def format_list_or_queryset(obj: list | QuerySet) -> str:
        ret = "["
        if len(obj) > 0:
            multiline = len(obj) > 1 or (len(obj) == 1 and not is_scalar(obj[0]))
            if multiline:
                ret += "\n"
            for value in obj:
                if multiline:
                    ret += " " * (current_depth * indent + indent)
                ret += simple_pformat(value, indent=indent, current_depth=current_depth + 1)
                if multiline:
                    ret += ",\n"
            if multiline:
                ret += " " * (current_depth * indent)
        ret += "]"
        return ret

    if isinstance(obj, dict):
        return format_dict(obj)
    if isinstance(obj, (list, QuerySet)):  # type: ignore
        return format_list_or_queryset(obj)
    return format_value(obj)


def soupify(value: str | bytes) -> BeautifulSoup:
    """
    Background: BeautifulSoup wrongly guessed the encoding of API json
    responses as latin-1, which lead to bastardized strings and much agony
    until I finally found out why. Always run soup-creation through this!
    """
    if isinstance(value, bytes):
        return BeautifulSoup(value, 'html.parser', from_encoding='utf-8')
    return BeautifulSoup(value, 'html.parser')


def time_querysets(*querysets: QuerySet, iterations=10, quiet=False):
    """Purely a testing function to be used in the CLI."""
    last_percent = 0
    measurements = []

    for i in range(iterations):
        start_time = time.time()
        for queryset in querysets:
            list(copy.deepcopy(queryset))
        elapsed_time = time.time() - start_time
        measurements.append(elapsed_time)
        if quiet:
            percent = int((i + 1) / iterations * 100)
            if not percent % 10 and percent != last_percent:
                last_percent = percent
                output = f"{percent}%"
                if percent == 100:
                    print(output)
                else:
                    print(output, end=" ... ", flush=True)
        else:
            print(f"[{i + 1}/{iterations}: {elapsed_time}")

    print(f"Mean:   {mean(measurements)}")
    print(f"Median: {median(measurements)}")


def timedelta_formatter(
    value: timedelta | float | int,
    short_format: bool = False,
    rounded: bool = False
) -> str:
    # If value is float or int, we suppose it's number of seconds:
    if isinstance(value, (int, float)):
        seconds = int(value)
    else:
        seconds = int(value.total_seconds())
    hours = int(seconds / 3600)
    seconds -= (hours * 3600)
    minutes = int(seconds / 60)
    seconds -= (minutes * 60)
    if rounded:
        if minutes > 30:
            hours += 1
        if seconds > 30:
            minutes += 1
    if short_format:
        time_str = ""
        if hours:
            time_str += "{}h".format(hours)
        if minutes and (not rounded or not hours):
            time_str += "{}m".format(minutes)
        if seconds and (not rounded or (not hours and not minutes)):
            time_str += "{}s".format(seconds)
        return time_str or "0s"
    time_list = []
    if hours:
        time_list.append(ngettext("%(hours)d hour", "%(hours)d hours", hours) % {"hours": hours})
    if minutes and (not rounded or not hours):
        time_list.append(ngettext("%(min)d min", "%(min)d min", minutes) % {"min": minutes})
    if seconds and (not rounded or (not hours and not minutes)):
        time_list.append(ngettext("%(sec)d sec", "%(sec)d sec", seconds) % {"sec": seconds})
    return ", ".join(time_list)


def _get_param_list(required_params: str = "", optional_params: str = "") -> list[str]:
    required_params = required_params.strip()
    optional_params = optional_params.strip()
    return (
        (required_params.split(" ") if required_params else []) +
        (optional_params.split(" ") if optional_params else [])
    )



import warnings
from typing import Any

from rest_framework import serializers
from rest_framework.compat import uritemplate
from rest_framework.exceptions import APIException
from rest_framework.fields import CharField, ChoiceField, EmailField, IntegerField, empty
from rest_framework.schemas.openapi import AutoSchema, SchemaGenerator as _SchemaGenerator
from rest_framework.schemas.utils import get_pk_description, is_list_view

from django.conf import settings
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import FieldDoesNotExist
from django.db import models
from django.db.models.fields import NOT_PROVIDED
from django.utils.encoding import force_str


class ExtraResponse:
    code: str
    key: str
    description: str
    example: Any

    def __init__(self, code: str | int, key: str, description: str, example: Any = None):
        self.code = str(code) if isinstance(code, int) else code
        self.key = key
        self.description = description
        self.example = example


class Extra404Response(ExtraResponse):
    def __init__(self, key='not_found', description='Object not found', example=None):
        example = example or {'error': ['Not found.']}
        super().__init__(code=404, key=key, description=description, example=example)


class BaseSchemaGenerator(_SchemaGenerator):
    def get_schema(self, request=None, public=False):
        schema = super().get_schema(request, public)
        schema.update({
            'security': self.get_security(),
            'tags': self.get_tags(),
            'servers': self.get_servers(),
        })
        if 'components' not in schema:
            schema['components'] = {}
        schema['components']['securitySchemes'] = {}
        for sec_def in schema.get('security', {}):
            for k in sec_def.keys():
                schema['components']['securitySchemes'][k] = self.get_security_scheme(k)
        return schema

    def get_servers(self):
        return [{'url': settings.ROOT_URL}]

    def get_security(self):
        """
        Returns a list of security requirement objects, as per:
        https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#securityRequirementObject

        Object names should be the class names of object names returned by
        a view's get_authenticators() method.

        Each object name should correspond to a scheme returned by
        self.get_security_scheme().
        """
        return []

    def get_security_scheme(self, name):
        """
        Returns a security scheme object, as per:
        https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#securitySchemeObject
        """
        return {}

    def get_tags(self):
        """
        Returns a list of tag objects, as per:
        https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#tagObject
        """
        return []


class BaseSchemaMixin:
    operation_description = ''
    operation_summary: str | None = None
    optional_response_fields: dict
    view: Any

    def __init__(self, tags=None, operation_id_base=None, component_name=None, **kwargs):
        if 'optional_response_fields' in kwargs:
            self.optional_response_fields = kwargs.pop('optional_response_fields')
        elif not hasattr(self, 'optional_response_fields'):
            self.optional_response_fields = {}
        for key, value in kwargs.items():
            setattr(self, key, value)
        super().__init__(tags=tags, operation_id_base=operation_id_base, component_name=component_name)  # type: ignore

    def collect_extra_responses(self, path, method):
        responses = {}
        response: dict
        for extra_response in self.get_extra_responses(path, method):
            if extra_response.code not in responses:
                response = {
                    'description': extra_response.description,
                    'content': {'application/json': {'examples': {}}},
                }
                responses[extra_response.code] = response
            else:
                response = responses[extra_response.code]
                if response['description']:
                    response['description'] = 'Multiple reasons'
                else:
                    response['description'] = extra_response.description
            if extra_response.example:
                response['content']['application/json']['examples'].update({
                    extra_response.key: {
                        'description': extra_response.description,
                        'value': extra_response.example,
                    }
                })
        return responses

    def get_action(self) -> str | None:
        if hasattr(self, 'view'):
            return getattr(self.view, 'action', None)
        return None

    def get_extra_responses(self, path, method) -> list[ExtraResponse]:
        responses = []

        if method in ('POST', 'PUT', 'PATCH'):
            missing_fields = {}
            non_null_fields = {}
            non_blank_fields = {}
            choice_fields = {}
            max_length_fields = {}
            min_length_fields = {}
            max_value_fields = {}
            min_value_fields = {}
            email_fields = {}
            serializer = self.view.get_serializer(path, method)
            for field_name, field in serializer.fields.items():
                if not field.read_only:
                    if field.required and method != 'PATCH':
                        missing_fields[field_name] = ['This field is required.']
                    if not field.allow_null:
                        non_null_fields[field_name] = ['This field may not be null.']
                    if isinstance(field, CharField) and not field.allow_blank:
                        non_blank_fields[field_name] = ['This field may not be blank.']
                    if isinstance(field, CharField):
                        if field.max_length is not None:
                            max_length_fields[field_name] = [
                                f'Ensure this field has no more than {field.max_length} characters.'
                            ]
                        if field.min_length is not None and field.min_length > 1:
                            min_length_fields[field_name] = [
                                f'Ensure this field has at least {field.min_length} characters.'
                            ]
                    if isinstance(field, IntegerField):
                        if field.min_value is not None:
                            min_value_fields[field_name] = [
                                f'Ensure this value is greater than or equal to {field.min_value}.'
                            ]
                        if field.max_value is not None:
                            max_value_fields[field_name] = [
                                f'Ensure this value is less than or equal to {field.max_value}.'
                            ]
                    if isinstance(field, ChoiceField):
                        choice_fields[field_name] = ['"foo" is not a valid choice.']
                    if isinstance(field, EmailField):
                        email_fields[field_name] = ['Enter a valid email address.']
            if missing_fields:
                responses.append(ExtraResponse(400, 'missing_fields', 'Missing fields', missing_fields))
            if non_null_fields:
                responses.append(ExtraResponse(400, 'non_null_fields', 'Non-nullable fields', non_null_fields))
            if non_blank_fields:
                responses.append(ExtraResponse(400, 'non_blank_fields', 'Non-blankable fields', non_blank_fields))
            if choice_fields:
                responses.append(ExtraResponse(400, 'invalid_choices', 'Invalid enum choices', choice_fields))
            if max_length_fields:
                responses.append(ExtraResponse(
                    400,
                    'max_length_exceeded',
                    'Maximum string length exceeded',
                    max_length_fields
                ))
            if min_length_fields:
                responses.append(ExtraResponse(
                    400,
                    'min_length_subceeded',
                    'Minimum string length subceeded',
                    min_length_fields
                ))
            if email_fields:
                responses.append(ExtraResponse(400, 'invalid_email', 'Invalid email address', email_fields))
            if min_value_fields:
                responses.append(ExtraResponse(
                    400,
                    'min_value_subceeded',
                    'Minimum integer value subceeded',
                    min_value_fields
                ))
            if max_value_fields:
                responses.append(ExtraResponse(
                    400,
                    'max_value_exceeded',
                    'Maximum integer value exceeded',
                    max_value_fields
                ))
        return responses

    def get_operation_summary(self, path, method):
        return self.operation_summary

    def get_optional_response_fields(self, path, method):
        fields = set()
        if method in self.optional_response_fields:
            fields.update(self.optional_response_fields[method])
        if 'default' in self.optional_response_fields:
            fields.update(self.optional_response_fields['default'])
        return list(fields)

    def get_path_parameter_description(self, path, method, variable) -> str | None:
        return None

    def get_path_parameter_type(self, path, method, variable):
        return 'string'

    def collect_extra_response_field_attributes(self, path, method, schema, field_name):
        schema.update(self.get_extra_response_field_attributes(path, method, field_name))
        properties = None
        if schema.get('type', '') == 'array' and 'items' in schema:
            if schema['items'].get('type', '') == 'object' and 'properties' in schema['items']:
                properties = schema['items']['properties']
        elif schema.get('type', '') == 'object' and 'properties' in schema:
            properties = schema['properties']
        if properties:
            for subname, subschema in properties.items():
                self.collect_extra_response_field_attributes(path, method, subschema, f'{field_name}.{subname}')

    def get_extra_response_field_attributes(self, path, method, field_name):
        """
        For use when we want to put some arbitrary attributes on a response
        field, like "deprecated".

        :param field_name: A "composite" fieldname, consisting of period
        separated fieldnames, e.g. `sections.file.url`.
        """
        return {}

    def get_security(self, path, method):
        return [{auth.__class__.__name__: []} for auth in self.view.get_authenticators()]

    def is_list_view(self, path, method):
        return is_list_view(path, method, self.view)

    def map_model_field(self, field):
        # Used to describe path parameters
        if isinstance(field, (models.AutoField, models.IntegerField)):
            schema = {'type': 'integer'}
        elif isinstance(field, models.BinaryField):
            schema = {'type': 'string', 'format': 'binary'}
        elif isinstance(field, models.BooleanField):
            schema = {'type': 'boolean'}
        elif isinstance(field, models.DateTimeField):
            schema = {'type': 'string', 'format': 'date-time'}
        elif isinstance(field, models.DateField):
            schema = {'type': 'string', 'format': 'date'}
        elif isinstance(field, models.TimeField):
            schema = {'type': 'string', 'format': 'time'}
        elif isinstance(field, (models.DecimalField, models.FloatField)):
            schema = {'type': 'number', 'format': 'float'}
        elif isinstance(field, models.URLField):
            schema = {'type': 'string', 'format': 'uri'}
        elif isinstance(field, models.EmailField):
            schema = {'type': 'string', 'format': 'email'}
        elif isinstance(field, models.UUIDField):
            schema = {'type': 'string', 'format': 'uuid'}
        else:
            schema = {'type': 'string'}
        if field.null:
            schema['nullable'] = True  # type: ignore
        return schema


class BaseSchema(BaseSchemaMixin, AutoSchema):
    operation_tags: list = []
    # Field schemas must contain "type" key. Key "format" is also common,
    # and we may want a "description".
    # https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.3.md#schema-object
    serializer_field_schemas: dict = {}

    def __init__(self, tags=None, operation_id_base=None, component_name=None, **kwargs):
        operation_id_base = operation_id_base or getattr(self, 'operation_id_base', None)
        super().__init__(tags=tags, operation_id_base=operation_id_base, component_name=component_name, **kwargs)

    def get_description(self, path, method):
        return self.operation_description or super().get_description(path, method)

    def get_operation(self, path, method):
        operation = super().get_operation(path, method)
        operation['summary'] = self.get_operation_summary(path, method) or operation['operationId']
        operation['security'] = self.get_security(path, method)
        return operation

    def get_operation_id(self, path, method):
        # It's ugly when not all ID's get uppercase first letter
        id_: str = super().get_operation_id(path, method)
        return id_[0].upper() + id_[1:]

    def get_path_parameters(self, path, method):
        # Had to override the whole thing to fix the check for model, get
        # custom descriptions & type, and check for model field type
        assert uritemplate, '`uritemplate` must be installed for OpenAPI schema support.'
        # My fix #1 (OK, somewhat messy):
        try:
            model = self.view.serializer_class.Meta.model
        except AttributeError:
            try:
                model = self.view.queryset.model
            except AttributeError:
                model = None

        parameters = []
        for variable in uritemplate.variables(path):
            # My fix #2:
            description = self.get_path_parameter_description(path, method, variable)
            schema = {'type': self.get_path_parameter_type(path, method, variable)}
            if model is not None:
                try:
                    model_field = model._meta.get_field(variable)
                except (FieldDoesNotExist, KeyError):
                    model_field = None
                if model_field is not None:
                    # Custom descriptions take precedence
                    if description is None and model_field.help_text:
                        description = force_str(model_field.help_text)
                    elif description is None and model_field.primary_key:
                        description = get_pk_description(model, model_field)
                    schema = self.map_model_field(model_field)
                    # Really makes no sense to mark path parameters as nullable
                    schema.pop('nullable', None)
            parameter = {
                'name': variable,
                'in': 'path',
                'required': True,
                'description': description or '',
                'schema': schema,
            }
            parameters.append(parameter)
        return parameters

    def get_request_body(self, path, method):
        # Default implementation just gets a reference and returns
        # {"schema": {"$ref": "#/components/schemas/..."}} for each content
        # type. However, this makes the request body identical for POST, PUT,
        # and PATCH, and we don't want that; for PATCH, _no_ parameters should
        # be marked as "required".
        if method not in ('PUT', 'PATCH', 'POST'):
            return {}
        self.request_media_types = self.map_parsers(path, method)
        serializer = self.get_serializer(path, method)
        if isinstance(serializer, serializers.ListSerializer):
            serializer = serializer.child
            is_list = True
        else:
            is_list = False
        if not isinstance(serializer, serializers.Serializer):
            item_schema = {}
        else:
            item_schema = self.map_serializer(serializer)
            if method == 'PATCH':
                item_schema['required'] = []
                for key, prop in item_schema.get('properties', {}).items():
                    if isinstance(prop, dict) and 'type' in prop:
                        if prop['type'] == 'object':
                            item_schema['properties'][key]['required'] = []
                        if prop['type'] == 'array' and 'items' in prop:
                            item_schema['properties'][key]['items']['required'] = []
        if is_list:
            request_schema = {
                'type': 'array',
                'items': item_schema,
            }
        else:
            request_schema = item_schema
        return {
            'content': {
                ct: {'schema': request_schema} for ct in self.request_media_types
            }
        }

    def get_responses(self, path, method):
        # Had to override the whole shit to inject specific response
        # descriptions
        is_list = False
        if method == 'DELETE':
            return {
                '204': {'description': 'Object successfully deleted'},
                **self.collect_extra_responses(path, method),
            }
        self.response_media_types = self.map_renderers(path, method)
        if self.is_list_view(path, method):
            serializer = self.get_serializer(path, method, many=True)
            is_list = True
        else:
            serializer = self.get_serializer(path, method)
        if isinstance(serializer, serializers.Serializer):
            item_schema = self.map_serializer(serializer)
        elif isinstance(serializer, serializers.ListSerializer) and \
                isinstance(serializer.child, serializers.BaseSerializer):
            item_schema = self.map_serializer(serializer.child)
            is_list = True
        else:
            item_schema = {}
        # Injected thing; all response fields are considered "required"
        # per default, put exceptions in self.optional_response_fields
        if 'properties' in item_schema:
            item_schema['required'] = [
                f for f in item_schema['properties'] if f not in self.get_optional_response_fields(path, method)]
            for name, schema in item_schema['properties'].copy().items():
                if 'writeOnly' in schema:
                    del item_schema['properties'][name]
                    continue
                if 'default' not in schema:
                    try:
                        if isinstance(serializer, serializers.ListSerializer):
                            meta = getattr(serializer.child, "Meta", None)
                        else:
                            meta = getattr(serializer, "Meta", None)
                        model_field = meta.model._meta.get_field(name) if meta is not None else None
                        if model_field and model_field.default != NOT_PROVIDED and not callable(model_field.default):
                            item_schema['properties'][name]['default'] = model_field.default
                    except (FieldDoesNotExist, AttributeError):
                        pass
                # Injected thing:
                self.collect_extra_response_field_attributes(path, method, schema, name)
        if is_list:
            response_schema = {
                'type': 'array',
                'items': item_schema,
            }
            paginator = self.get_paginator()
            if paginator:
                response_schema = paginator.get_paginated_response_schema(response_schema)  # type: ignore
        else:
            response_schema = item_schema
        if method == 'POST':
            code = '201'
        else:
            code = '200'
        return {
            code: {
                'content': {
                    ct: {'schema': response_schema}
                    for ct in self.response_media_types
                },
                'description': ''
            },
            **self.collect_extra_responses(path, method),
        }

    def get_serializer(self, path, method, many=False):
        # Mock Request object because some serializers depend on method
        class MockRequest:
            method: str
            user: AnonymousUser

        self.view.request = MockRequest()
        self.view.request.method = method
        self.view.request.user = AnonymousUser()

        if many:
            try:
                return self.view.get_serializer(many=many)
            except (APIException, AttributeError):
                warnings.warn(
                    f'{self.view.__class__.__name__}.get_serializer() raised an exception during schema generation. '
                    f'Serializer fields will not be generated for {method} {path}.'
                )
                return None
        return super().get_serializer(path, method)

    def get_tags(self, path, method):
        return [t.name for t in self.operation_tags]

    def map_choicefield(self, field):
        # Patch to handle TextChoices and IntegerChoices objects
        mapping = super().map_choicefield(field)
        if 'type' in mapping:
            if mapping['type'] == 'string':
                mapping['enum'] = [str(v) for v in mapping['enum']]
            elif mapping['type'] == 'integer':
                mapping['enum'] = [int(v) for v in mapping['enum']]
        return mapping

    def map_field(self, field):
        field_schema: dict = {}

        # Overriding to check for pk_field and PK model field types
        if isinstance(field, serializers.PrimaryKeyRelatedField):
            model = getattr(field.queryset, 'model', None)
            if model is not None:
                field_schema = self.map_model_field(model._meta.pk)
            elif field.pk_field is not None:
                field_schema = self.map_field(field.pk_field)  # type: ignore

        if not field_schema:
            field_schema = super().map_field(field)
        if field.field_name in self.serializer_field_schemas:
            field_schema.update(self.serializer_field_schemas[field.field_name])
        if field.default and field.default != empty and not callable(field.default):
            # This is also done in map_serializer(), but I have to do it here
            # so I can also do the check for model field default below
            field_schema['default'] = field.default
        else:
            try:
                meta = getattr(field.parent, "Meta", None)
                model_field = meta.model._meta.get_field(field.field_name) if meta is not None else None
                if model_field and model_field.default != NOT_PROVIDED and not callable(model_field.default):
                    field_schema['default'] = model_field.default
            except (FieldDoesNotExist, AttributeError):
                pass
        return field_schema


### MIXINS #####################################

class ModelViewMixin(BaseSchemaMixin):
    object_name: str
    object_name_plural: str
    view: Any

    def get_operation_summary(self, path, method):
        if self.operation_summary is not None:
            return self.operation_summary
        object_name = getattr(self, 'object_name', None)
        object_name_plural = getattr(self, 'object_name_plural', None)
        if not object_name_plural:
            object_name_plural = object_name
        if not object_name:
            serializer = self.get_serializer(path, method)  # type: ignore
            if hasattr(serializer, 'Meta'):
                if not object_name:
                    object_name = serializer.Meta.model._meta.verbose_name
                if not object_name_plural:
                    object_name_plural = serializer.Meta.model._meta.verbose_name_plural
            else:
                return None
        if method == 'GET':
            if self.is_list_view(path, method):
                return f'List {object_name_plural}'
            return f'Get {object_name}'
        if method == 'POST':
            return f'Create {object_name}'
        if method == 'PUT':
            return f'Update {object_name}'
        if method == 'PATCH':
            return f'Patch {object_name}'
        if method == 'DELETE':
            return f'Delete {object_name}'
        return None

    def get_404_response(self, path, method):
        """Refers to the object fetched by view.get_object()"""
        serializer = self.get_serializer(path, method)  # type: ignore
        example = None
        object_name = self.view.get_queryset().model._meta.object_name
        if not object_name and hasattr(serializer, 'Meta') and serializer.Meta.model:
            object_name = serializer.Meta.model._meta.object_name
        if object_name:
            example = {'error': [f'No {object_name} matches the given query.']}
        return Extra404Response(example=example)

    def get_extra_responses(self, path, method):
        responses = super().get_extra_responses(path, method)
        if method == 'POST' or (method == 'GET' and self.is_list_view(path, method)):
            return responses
        _404_response = self.get_404_response(path, method)
        if _404_response:
            responses.append(_404_response)
        return responses



from django import forms
from django.core.exceptions import ValidationError
from django.forms.utils import ErrorDict
from django.utils.translation import gettext as _


class CustomEmailField(forms.EmailField):
    def run_validators(self, value):
        """
        Had to override this whole thing just to be able to insert `value`
        into the error message. :-/
        """
        if value in self.empty_values:
            return
        errors = []
        for v in self.validators:
            try:
                v(value)
            except ValidationError as e:
                if hasattr(e, 'code'):
                    if e.code == "invalid":
                        e.message = _("'%(email)s' is not a valid email address.") % {"email": value}
                    elif e.code in self.error_messages:
                        e.message = self.error_messages[e.code]
                errors.extend(e.error_list)
        if errors:
            raise ValidationError(errors)


class ErrorDictFormSet(forms.BaseFormSet):
    """
    Adds utility method get_error_dict(), which converts self.errors (normally
    a list) to an ErrorDict, where the keys are (hopefully) the same as the
    "name" attributes of the form elements in the rendered formset.
    """
    def get_error_dict(self) -> ErrorDict:
        error_dict = ErrorDict()
        if self.non_form_errors():
            error_dict.update({"__all__": self.non_form_errors()})
        for idx, form_errors in enumerate(self.errors):
            assert isinstance(form_errors, ErrorDict)
            error_dict.update({
                "%s-%s" % (self.add_prefix(idx), k): v
                for k, v in form_errors.items()
            })
        return error_dict


class ErrorDictModelFormSet(ErrorDictFormSet, forms.BaseModelFormSet):
    ...



from dataclasses import MISSING, InitVar, dataclass, field, fields
from typing import Any, Literal

from django.conf import settings
from django.utils.functional import empty


class MissingSettingsError(Exception):
    keys: list[str]

    def __init__(self, members: str | list[str] | list["MissingSettingsError"], *args):
        print(args)
        super().__init__(*args)
        keys = []
        if not isinstance(members, list):
            members = [members]
        for member in members:
            if isinstance(member, MissingSettingsError):
                keys.extend(member.keys)
            else:
                keys.append(member)
        self.keys = keys

    def __str__(self):
        return ", ".join(self.keys)


@dataclass
class SettingsBase:
    user_settings: InitVar[Any]

    def __post_init__(self, user_settings: Any):
        errors = []

        for own_field in fields(self):
            if isinstance(user_settings, dict):
                user_setting = user_settings.get(own_field.name, empty)
            elif user_settings is not None:
                user_setting = user_settings
            else:
                user_setting = empty

            if isinstance(own_field.type, type) and issubclass(own_field.type, SettingsBase):
                try:
                    setattr(self, own_field.name, own_field.type(user_settings=user_setting))
                except MissingSettingsError as e:
                    errors.append(MissingSettingsError([f"{own_field.name}.{key}" for key in e.keys]))
            elif user_setting is not empty:
                setattr(self, own_field.name, user_setting)
            else:
                if own_field.default is MISSING and own_field.default_factory is MISSING:
                    errors.append(MissingSettingsError(own_field.name))
        if errors:
            raise MissingSettingsError(errors)


@dataclass
class RunServerSettings(SettingsBase):
    DEFAULT_ADDR: str = field(init=False, default="127.0.0.1")
    DEFAULT_PORT: str = field(init=False, default="8000")
    SERVER: Literal["django", "daphne"] = field(init=False, default="django")


@dataclass
class Settings(SettingsBase):
    RUNSERVER: RunServerSettings = field(init=False)


gd_settings = Settings(getattr(settings, "KLAATU_DJANGO", None))



import operator
from abc import abstractmethod
from enum import Enum, auto
from pathlib import Path
from typing import Collection, Generic, TypeVar

from django.core.files.storage import FileSystemStorage as BaseFileSystemStorage, Storage


class SortKey(Enum):
    NAME = auto()
    SIZE = auto()
    ISDIR = auto()
    CTIME = auto()
    MTIME = auto()


SortKeyWithOrder = tuple[SortKey, bool]
SortKeyArg = Collection[SortKey | SortKeyWithOrder]
_T = TypeVar("_T", bound=object)


def normalize_sort_keys(keys: SortKeyArg) -> list[SortKeyWithOrder]:
    ret = []
    for key in keys:
        if isinstance(key, SortKey):
            ret.append((key, False))
        else:
            ret.append(key)
    return ret


class SortableStorage(Storage, Generic[_T]):
    def sort_by(self, files: Collection[_T], keys: SortKeyArg) -> list[_T]:
        nkeys = normalize_sort_keys(keys)
        reverse_all = False

        for (key, reverse) in nkeys:
            if key == SortKey.NAME:
                reverse_all = reverse

        return sorted(files, key=lambda f: self.get_sort_value(f, nkeys, reverse_all), reverse=reverse_all)

    @abstractmethod
    def get_sort_value(
        self,
        file: _T,
        nkeys: list[SortKeyWithOrder],
        reverse_all: bool
    ) -> list[str | int | float]:
        raise NotImplementedError("SortableStorage subclasses must implement get_sort_value().")

    @abstractmethod
    def list_recursive(self, path: str = "", sort: SortKeyArg | None = None) -> list[str]:
        """Recursively lists files in `path`, ordered by `sort`."""
        raise NotImplementedError("SortableStorage subclasses must implement list_recursive().")


class FileSystemStorage(SortableStorage[Path], BaseFileSystemStorage):
    def get_sort_value(self, file, nkeys, reverse_all):
        stat = file.stat()
        result: list[str | int | float] = []
        for (key, reverse) in nkeys:
            reverse = operator.xor(reverse, reverse_all)
            multiplier = -1 if reverse else 1
            if key == SortKey.NAME:
                result.append(file.name)
            elif key == SortKey.ISDIR:
                result.append(operator.xor(file.is_dir(), reverse))
            elif key == SortKey.SIZE:
                if file.is_dir():
                    result.append(0)
                else:
                    result.append(stat.st_size * multiplier)
            elif key == SortKey.CTIME:
                result.append(stat.st_ctime * multiplier)
            elif key == SortKey.MTIME:
                result.append(stat.st_mtime * multiplier)
        return result

    def recurse(self, path: Path | str, sort: SortKeyArg | None = None) -> list[Path]:
        """
        Recursively gets Path objects for files in `path`, ordered by `sort`.
        """
        files = []
        if isinstance(path, str):
            path = Path(path)
        for entry in path.iterdir():
            if entry.is_dir():
                files.extend(self.recurse(entry))
            else:
                files.append(entry)
        return self.sort_by(files, sort or [])

    def list_recursive(self, path="", sort=None):
        full_path = self.path(path)
        return [str(p)[len(full_path):].strip("/") for p in self.recurse(full_path, sort or [])]



from rest_framework.permissions import BasePermission
from rest_framework.request import Request

from django.contrib.auth.models import AbstractUser


class UserObjectPermissions(BasePermission):
    """
    Usage: Implement the class method `has_permission(user, verb)` and/or the
    object method `has_object_permission(user, verb)` on the models.

    No need to check superuser status in those methods, since it's done
    here in has_permission().

    Also, if the object implements `has_object_permission()`, we assume the
    user is required to be authenticated, so no need to check that either.
    """
    VERB_MAPPING = {
        'GET': 'view',
        'POST': 'create',
        'PUT': 'change',
        'PATCH': 'change',
        'DELETE': 'delete',
    }

    def is_superuser(self, request: Request):
        return isinstance(request.user, AbstractUser) and request.user.is_superuser

    def has_permission(self, request, view):
        get_queryset = getattr(view, "get_queryset", None)
        if self.is_superuser(request) or get_queryset is None:
            return True
        try:
            queryset = get_queryset()
        except AssertionError:
            # This is _probably_ because the view does not use a queryset
            return True
        if (
            request.method in self.VERB_MAPPING and
            hasattr(queryset, 'model') and
            hasattr(queryset.model, 'has_permission')
        ):
            return queryset.model.has_permission(request.user, verb=self.VERB_MAPPING[request.method])
        return True

    def has_object_permission(self, request, view, obj):
        if self.is_superuser(request):
            return True
        if obj is not None and request.method in self.VERB_MAPPING and hasattr(obj, 'has_object_permission'):
            if request.user.is_authenticated:
                return obj.has_object_permission(request.user, verb=self.VERB_MAPPING[request.method])
            return False
        return True



from django.contrib.staticfiles.management.commands.collectstatic import Command as BaseCommand


class Command(BaseCommand):
    def log(self, msg: str, level: int = 2):
        """
        MAJESTICALLY ugly hack to avoid all the annoying "Found another file"
        messages that are an unavoidable result of running Grappelli.
        """
        if not msg.startswith('Found another file with the destination path'):
            super().log(msg, level=level)



from django import forms


class CustomCheckboxSelectMultiple(forms.CheckboxSelectMultiple):
    """
    Renders a multiple checkbox select using Bootstrap's custom control layout:
    https://getbootstrap.com/docs/4.6/components/forms/#checkboxes-and-radios-1

    `field_wrapper_class` is for a <div> wrapping the entire widget.
    `option_wrapper_class` is for <div>'s wrapping each individual checkbox.

    Responsive layout example:

    CustomCheckboxSelectMultiple(
        field_wrapper_class="row",
        option_wrapper_class="col-12 col-lg-6"
    )
    """
    template_name = "klaatu/widgets/custom_checkbox_select.html"
    field_wrapper_class = ""
    option_wrapper_class = ""

    def __init__(self, attrs=None, choices=(), field_wrapper_class="", option_wrapper_class=""):
        super().__init__(attrs, choices)
        self.field_wrapper_class = field_wrapper_class
        self.option_wrapper_class = option_wrapper_class

    def create_option(self, name, value, label, selected, index, subindex=None, attrs=None):
        option = super().create_option(name, value, label, selected, index, subindex, attrs)
        if "class" in option["attrs"]:
            option["attrs"]["class"] += " form-check-input"
        else:
            option["attrs"]["class"] = "form-check-input"
        return option

    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)
        context["widget"].update(
            field_wrapper_class=self.field_wrapper_class,
            option_wrapper_class=self.option_wrapper_class,
        )
        return context



from rest_framework.renderers import BrowsableAPIRenderer as BaseBrowsableAPIRenderer


class BrowsableAPIRenderer(BaseBrowsableAPIRenderer):
    def get_filter_form(self, data, view, request):
        """
        Just to get rid of those pesky duplicate DB hits that make it
        harder to debug and optimize.
        """
        return None
